# Writing an HTTP Server in Cython+

[Cython+](https://www.cython.plus) is an ongoing project to extend [Cython](https://cython.org/) with [high performance GIL-free types](https://www.youtube.com/watch?v=Lb6q1hhBp-k) using static typing.

Our [next steps](https://www.cython.plus/P-CYP-Roadmap) include with writing a minimal standard library.

To drive this effort I'm writing a very basic HTTP server in Cython+, building everything I will need along the way.


## Usage

The repository looks like this:

```
- README.md
- src/
    - Makefile
    - setup.py
    - stdlib/
        - *.pxd
        - ...
    - *.pyx
    - ...
- ...
```

Each `.pyx` file in `src` is a standalone demo.

To run the demos (e.g. `helloworld.pyx`):

```
cd src
make helloworld
```


## GIL-free strings

So far Cython+ merely relied on C++'s `std::string` to represent strings. As it happens, writing our own GIL-free string implementation is one of our next steps.

As I started thinking about how to wrap the C socket API, I realised that it would go much easier with our own strings. In particular, C APIs often accept `null` as a valid value for `char *` string arguments, but there is no good way to represent `null` with `std::string`.

So I wrote a minimal string implementation by wrapping a `std::string` into a Cython+ GIL-free type. There are multiple advantages:
- Aliasing a string will not create a copy, unlike `std::string`
- We can write our own methods, e.g. to mirror Python's strings
- We can enforce immutability if we want

Implemented string methods include:
- `find`
- `substr` 	(plays the role of Python's slice operator)
- `__add__`	(string concatenation using `+`)
- `split`
- `join`

Since Cython+ GIL-free types are allocated to the heap and referenced by pointer, `null` is a valid value.

Actually at first I wanted to avoid `std::string` because they incur unneeded but unavoidable zeroing or copying overhead when interacting with C I/O APIs. This is very well explained in [the `basic_string::resize_and_overwrite` proposal for C++23](http://www.open-std.org/JTC1/SC22/WG21/docs/papers/2021/p1072r7.html). So the first version actually used `std::string_view` (C++17) with hand-crafted memory allocation using `malloc` and `free`. But that was incompatible with some C++ string libraries like the excellent [fmt](https://fmt.dev/), and that's why I went back to using `std::string` as backend.

That initial implementation remains available in branch `string`.


## Echo Server and Client

Using [Beej's Guide to Network Programming](https://beej.us/guide/bgnet/html//index.html) as a handy reference, I wrote adapters for needed parts of the POSIX socket API. As much as practical, these adapters mirror [Python's socket API](https://docs.python.org/3/library/socket.html).

These adapters include the functions:
- `getaddrinfo`
- `socket`
- `setsockopt`	(only for options with integer values, which is all I needed)
- `connect`
- `bind`
- `listen`
- `accept`
- `recv`
- `send`
- `shutdown`
- `close`

The new GIL-free strings proved very useful for this :)

Since currently Cython+ doesn't have GIL-free exceptions (also in our next steps), I take the GIL and raise Python exceptions instead when there is an error.

It's not complete, but it's enough to write a simple echo server and client demo:

```
make echoserver
```

and

```
make client
```


## String formatting

In order to (much) more easily compose response messages, I chose to implement string formatting. This provides a way to efficiently pack miscellaneous values into strings without repeated string concatenations. In fact, before this I had no easy way to even convert integer values to strings!

I chose to go with [fmt](https://fmt.dev), a neat C++ string formatting library. I discarded the string formatting solutions from the C standard library like `sprintf` because they (a) do not manage memory alllocation and instead require a buffer of sufficient size be allocated in advance to store the formatted string, and (b) they are not type safe. Since C++20 there is also a string formatting library in the C++ standard library, but I deemed that it would be easier to depend on an external library like `fmt` than to require such a recent version of C++.

To install `fmt` you can follow the explanations in the [usage guide](https://fmt.dev/latest/usage.html). It must be built as a **static library** because we use it in Python extensions; the guide explains how (tldr: `cmake -DCMAKE_POSITION_INDEPENDENT_CODE=TRUE ...`).

Note that since `fmt` makes use of C++ templates and compile-time assertions, Cython will not always be able to catch misuses and the user can be confronted with C++ compilation errors of the underlying compilation step. This is still much better than misuses not being caught at all and resulting in runtime faults as can happen with the C library `sprintf`!

I wrote an adapter for the `fmt::format` method so that our GIL-free strings may be accepted as arguments and be type of the returned value, instead of just `std::string`. This proved rather complicated and I had to come up with several workaround hacks:

1. First, Cython and Cython+ have limited support for C++ templates, so I had to write the adapter directly in C++ using Cython's foreign code interface.
2. But such raw C++ foreign code is inserted before user-defined types such as our GIL-free string class, so I had to use templates that can be instantiated with our string type later in the code.
3. I then overrode the `cname` (the name in the generated C++ code) of my adapter template method to explictly instantiate it with our GIL-free string type.
4. To this end I needed the `cname` of our string type, but in Cython user-defined names are normally mangled, so I had to override its `cname` as well just to know what it would be.

Anyway, it works!

I kept the format string argument as a simple `char *` because the prevailing use case by far is to just pass a string literal.

The formatting syntax is the same as the underlying `fmt` library. It's quite similar to [Python's Format Specification Mini Language](https://docs.python.org/3.8/library/string.html#formatspec), and in fact it's inspired by it.


## A Simple HTTP/1.0 Echo Server

With all this in hand, I wrote a minimal HTTP/1.0 server. It's a kind of echo server that parses HTTP requests and responds with the same body as the message you send it.

I started by writing a simple HTTP/1.0 parser using basic string operations such as `find` and `substr`. While I was at it I wrote a URL parser as well.

The server waits until it has received all the HTTP headers (delimited by a double `\r\n`). If the headers specify a `Content-Length`, it waits again until the whole message body is received, and then responds with `200 OK` and the same message body. Otherwise it just sends a `200 OK` response. If it fails to parse the request, it answers with [`418 I'm a teapot`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/418).

To start the server:

```
$ make httpserver
>>> import httpserver
listening on http://0.0.0.0:3490
```

You can use `wget` to test it:

```
wget http://0.0.0.0:3490
```

To send data:

```
wget http://0.0.0.0:3490 --post-data="Hello!"
```

or

```
wget http://0.0.0.0:3490 --post-file=helloworld.pyx
```

To stop the server, you can send `quit`:

```
wget http://0.0.0.0:3490 --post-data="quit"
```

You can also send non-HTTP data using the simple client I wrote earlier, and stop the server just by sending `quit`.

